#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Copyright (c) 2017 Alan Carter. All rights reserved.
#

################################################################################
# Imports
################################################################################
import json
import requests
import urllib2
import time
import datetime
import re
import sys
import indigoPluginUpdateChecker

################################################################################
# Globals
################################################################################

########################################
def updateVar(name, value, folder=0):
    if name not in indigo.variables:
        indigo.variable.create(name, value=value, folder=folder)
    else:
        indigo.variable.updateValue(name, value)

################################################################################
class Plugin(indigo.PluginBase):
    ########################################
    # Class properties
    ########################################
    
    ########################################
    def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs): 
        indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
        self.userName = pluginPrefs.get("userName", "None")
        self.userPass = pluginPrefs.get("userPass", "None")
        self.clientId = pluginPrefs.get("clientId", "None")
        self.clientSecret = pluginPrefs.get("clientSecret", "None")
        self.accessToken = pluginPrefs.get("accessToken", "None")
        self.refreshToken = pluginPrefs.get("refreshToken", "None")
        self.expiration = int(pluginPrefs.get("expiration", 0))
        self.welcome = pluginPrefs.get("Welcome", False)
        self.pollFreq = float(pluginPrefs.get("netatmoFreq", 1))
        self.pollF = int(self.pollFreq * 60)
        self.loggingCam = pluginPrefs.get("LogChoiceCam", False)
        self.loggingMon = pluginPrefs.get("LogChoiceMon", False)
        self.loggingWs = pluginPrefs.get("LogChoiceWs", False)
        self.processLastSeen = pluginPrefs.get("ProcessLastSeen", False)
        self.delForgotten = pluginPrefs.get("DeleteForgotten", False)
        self.snapshotPath = pluginPrefs.get("SnapshotPath", "")
        self.healthyHome = pluginPrefs.get("HealthyHome", False)
        self.weather = pluginPrefs.get("Weather", False)
        self.syncCamNames = pluginPrefs.get("SyncCameraNames", False)
        self.syncMonNames = pluginPrefs.get("SyncMonitorNames", False)
        self.showDirectionInNotes = pluginPrefs.get("ShowDirectionInNotes", False)
        self.tUnits = (" °C"," °F")
        self.rUnits = (" mm", " in")
        self.wUnits = (" kph", " mph", " ms", " Bft", " knot")
        self.bftThreshold = (1, 5, 11, 19, 28, 38, 49, 61, 74, 88,102, 117)
        self.wConversion = (1, 0.621371, 0.277778, 0, 0.539957)
        self.pUnits = (" mbar", " inHg", "mmHg")
        self.homeId = ""
        self.camUrl = {}
        self.camId = {}
        self.personId = {}
        self.currentPersonsList = []
        self.commsOK = True
        self.retriesC = 0
        self.retriesH = 0
        self.retriesW = 0
        self.lastUpdateTimeHH = 0
        self.lastUpdateTimeWS = 0
        self.lastCO2 = {}
        self.lastRain = 0
        self.lastRainH = 0
        self.lastRainD = 0
        self.updater = indigoPluginUpdateChecker.updateChecker(self, "https://www.dropbox.com/s/zzwrr97wwqo0dqg/versionInfo.html?dl=1", 7)

    ########################################
    def __del__(self):
        indigo.PluginBase.__del__(self)
        
    ########################################
    def startup(self):
        indigo.server.log("Starting Netatmo Multi plugin")
        self.updater.checkVersionPoll()
        if self.welcome or self.healthyHome or self.weather:
            self.createDevices()
        else:
            indigo.server.log("Please enable at least one Netatmo device in Config", isError=True)
            return

    def deviceStartComm(self, dev):
        dev.stateListOrDisplayStateIdChanged()
        return

    def shutdown(self):
        indigo.server.log("Stopping Netatmo Multi plugin")
        self.pluginPrefs["accessToken"] = self.accessToken
        self.pluginPrefs["refreshToken"] = self.refreshToken
        self.pluginPrefs["expiration"] = self.expiration
        
        
    ########################################
    def runConcurrentThread(self):
        indigo.server.log("********************************************************************")
        indigo.server.log("Starting Netatmo Multi monitoring thread")
        if self.welcome:
            indigo.server.log("Welcome camera(s) will be polled")
            indigo.server.log(" - Polling set to %s seconds" % self.pollF)
            if self.processLastSeen:
                indigo.server.log(" - Last Seen variables will be created/updated")
            else:
                indigo.server.log(" - Last Seen variables will not be created/updated")
            if self.delForgotten:
                indigo.server.log(" - Persons forgotten in Netatmo app will be deleted")
            else:
                indigo.server.log(" - Persons forgotten in Netatmo app will not be deleted")
            if self.snapshotPath != "":
                indigo.server.log(" - Snapshot folder is %s" % self.snapshotPath)
        if self.healthyHome:
            indigo.server.log("Healthy Home air quality monitor(s) will be polled")
        if self.weather:
            indigo.server.log("Weather station(s) will be polled")
        if self.loggingCam:
            indigo.server.log("Logging all Welcome camera messages to Event Log")
        else:
            indigo.server.log("Welcome camera message logging is off")
        if self.loggingMon:
            indigo.server.log("Logging all Healthy Home messages to Event Log")
        else:
            indigo.server.log("Healthy Home message logging is off")
        if self.loggingWs:
            indigo.server.log("Logging all weather station messages to Event Log")
        else:
            indigo.server.log("Weather station message logging is off")
        indigo.server.log("********************************************************************")
        try:
            while True:
                if self.expiration < time.time():   # Token should be renewed
                    status, detail = self.refresh(self.clientId, self.clientSecret, self.refreshToken)
                    if status == "error" and self.commsOK:
                        indigo.server.log("runConcurrentThread() - server comms failed.  Will keep trying.", isError=True)
                        self.commsOK = False
                    elif status != "error" and self.commsOK == False:
                        indigo.server.log("runConcurrentThread() - server comms now OK")
                        self.commsOK = True
                if self.commsOK:
                    self.getUpdate()
                    self.updater.checkVersionPoll()
                else:
                    if self.welcome:
                        device = None
                        for dev in indigo.devices.iter("self"):
                            if dev.address == "NW0":
                                device = indigo.devices[dev.name]
                        if device != None:
                            device.updateStateOnServer("status", value = "off")
                    self.sleep(600)
                    self.commsOK = True
                                
                self.sleep(self.pollF)
                
        except self.StopThread:
            self.shutdown()


    ######################################
    # Called when user prefs are changed   
            
    def closedPrefsConfigUi(self, valuesDict, userCancelled):
        # If the user saved preferences, update logging parameters
        if userCancelled == False:
            oldLoggingCam = self.loggingCam
            self.loggingCam = valuesDict.get("LogChoiceCam", False)
            if self.loggingCam != oldLoggingCam:
                if self.loggingCam:
                    indigo.server.log("Logging all Welcome camera messages to Event Log")
                else:
                    indigo.server.log("Welcome camera message logging is off")
            oldSyncCam = self.syncCamNames
            self.syncCamNames = valuesDict.get("SyncCameraNames", False)
            if self.syncCamNames != oldSyncCam:
                if self.syncCamNames:
                    indigo.server.log("Camera name changes will be synced to Indigo device names")
                else:
                    indigo.server.log("Camera name changes will not be synced to Indigo device names")
            oldLoggingMon = self.loggingMon
            self.loggingMon = valuesDict.get("LogChoiceMon", False)
            if self.loggingMon != oldLoggingMon:
                if self.loggingMon:
                    indigo.server.log("Logging all Healthy Home messages to Event Log")
                else:
                    indigo.server.log("Healthy Home message logging is off")
            oldSyncMon = self.syncMonNames
            self.syncMonNames = valuesDict.get("SyncMonitorNames", False)
            if self.syncMonNames != oldSyncMon:
                if self.syncMonNames:
                    indigo.server.log("Healthy Home name changes will be synced to Indigo device names")
                else:
                    indigo.server.log("Healthy Home name changes will not be synced to Indigo device names")
            oldLoggingWs = self.loggingWs
            self.loggingWs = valuesDict.get("LogChoiceWs", False)
            if self.loggingWs != oldLoggingWs:
                if self.loggingWs:
                    indigo.server.log("Logging all weather station messages to Event Log")
                else:
                    indigo.server.log("Weather station message logging is off")
            oldLast = self.processLastSeen  
            self.processLastSeen = valuesDict.get("ProcessLastSeen", False)
            if self.processLastSeen != oldLast:
                if self.processLastSeen:
                    indigo.server.log("Last Seen variables will be created/updated")
                else:
                    indigo.server.log("Last Seen variables will not be created/updated")
            oldDel = self.delForgotten
            self.delForgotten = valuesDict.get("DeleteForgotten", False)
            if self.delForgotten != oldDel:
                if self.delForgotten:
                    indigo.server.log("Persons forgotten in Netatmo app will be deleted")
                else:
                    indigo.server.log("Persons forgotten in Netatmo app will not be deleted")
            oldSp = self.snapshotPath
            self.snapshotPath = valuesDict.get("SnapshotPath", "")
            if self.snapshotPath != oldSp:
                if self.snapshotPath == "":
                    indigo.server.log("No snapshot path specified", isError=True)
                else:
                    indigo.server.log("New snapshot folder path is %s" % self.snapshotPath)  
            oldHh = self.healthyHome
            self.healthyHome = valuesDict.get("HealthyHome", False)
            if self.healthyHome != oldHh:
                if self.healthyHome:
                    indigo.server.log("Healthy Home monitor(s) will be polled")
                    indigo.server.log("*** If you have added a monitor you must restart the plugin to create a new Indigo device ***")
                else:
                    indigo.server.log("Healthy Home monitor(s) will not be polled")
            oldWs = self.weather
            self.weather = valuesDict.get("Weather", False)
            if self.weather != oldWs:
                if self.weather:
                    indigo.server.log("Weather station will be polled")
                    indigo.server.log("*** If you have added a module you must restart the plugin to create a new Indigo device ***")
                else:
                    indigo.server.log("Weather station(s) will not be polled")
            oldSd = self.showDirectionInNotes
            self.showDirectionInNotes = valuesDict.get("ShowDirectionInNotes", False)
            if self.showDirectionInNotes != oldSd:
                if self.showDirectionInNotes:
                    indigo.server.log("Wind direction will be shown in Indigo Home Window Notes field")
                else:
                    indigo.server.log("Wind direction will not be shown in Indigo Home Window Notes field")

        
    ########################################
    # Netatmo specific functions
    
    def createDevices(self):
        if self.userName == "None":
            indigo.server.log("Missing User Name")
        else:
            pass    
        if self.userPass == "None":
            indigo.server.log("Missing User Password")
        else:
            pass    
        if self.clientId == "None":
            indigo.server.log("Missing Client ID")
        else:
            pass    
        if self.clientSecret == "None":
            indigo.server.log("Missing Client Secret")
        else:
            pass
        if (self.accessToken == "") or (self.refreshToken == ""):
            status, detail = self.authenticate(self.userName, self.userPass, self.clientId, self.clientSecret)
            if status == "error" and self.commsOK:
                indigo.server.log("Authentication failed in createDevices", isError=True)
                indigo.server.log(detail, isError=True)
                self.commsOK = False
                #return
            elif status == "OK" and self.commsOK == False:
                indigo.server.log("Authentication in createDevices now OK")
                self.commsOK = True
        else:
        	indigo.server.log("Current tokens are valid; authentication not required")
        
        if self.commsOK:
            if self.welcome:        
                if not ("Netatmo" in indigo.variables.folders):
                    indigo.variables.folder.create("Netatmo")   
                self.netatmoFolderId = indigo.variables.folders.getId("Netatmo")
                payload = {'access_token': self.accessToken,'size': 5,}
                try:
                    response = requests.post("https://api.netatmo.com/api/gethomedata", data=payload)
                    response.raise_for_status()
                    if response.status_code == 403:  # This code doesn't work
                        self.refresh(self.clientId, self.clientSecret, self.refreshToken)
                    else:
                        dataJ = response.json()["body"]
                        self.homeId = dataJ["homes"][0]["id"]
                        max_cameras = len(dataJ["homes"][0]["cameras"])
                    #indigo.server.log("Max Cams = "+str(max_cameras))
                    for cam in range(0, max_cameras):
                        camName = str(dataJ["homes"][0]["cameras"][cam]["name"])
                        camStatus = str(dataJ["homes"][0]["cameras"][cam]["status"])
                        if camStatus == "on":
                        	self.camUrl[cam] = str(dataJ["homes"][0]["cameras"][cam]["vpn_url"])
                        else:
                        	self.camUrl[cam] = ""
                        self.camId[cam] = str(dataJ["homes"][0]["cameras"][cam]["id"])
                        camAdd = "NW%s" % cam
                        device = None
                        for dev in indigo.devices.iter("self"):
                            if dev.address == camAdd:
                                device = indigo.devices[dev.name]
                        if device == None:
                            indigo.server.log("Creating Netatmo camera device for %s" % (camName))
                            nameC = "%s Camera" % (camName)
                            device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=camAdd, name=nameC, pluginId="com.racarter.indigoplugin.netatmo-welcome", deviceTypeId="netatmoWelcome", props={})
                except requests.exceptions.HTTPError as error:
                   indigo.server.log("Camera device creation error: "+self.getErrorText(error.response.text), isError=True)
                   
            if self.healthyHome:
                payload = {'access_token': self.accessToken}
                try:
                    response = requests.post("https://api.netatmo.com/api/gethomecoachsdata", params=payload)
                    response.raise_for_status()
                    dataJ = response.json()["body"]
                    max_monitors = len(dataJ["devices"])
                    for mon in range(0, max_monitors):
                        monName = str(dataJ["devices"][mon]["station_name"])
                        monAdd = "HH%s" % mon
                        device = None
                        for dev in indigo.devices.iter("self"):
                            if dev.address == monAdd:
                                device = indigo.devices[dev.name]
                        if device == None:
                            indigo.server.log("Creating Netatmo monitor device for %s" % (monName))
                            nameM = "%s Air Monitor" % (monName)
                            device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=monAdd, name=nameM, pluginId="com.racarter.indigoplugin.netatmo-welcome", deviceTypeId="netatmoHome", props={})         
                except requests.exceptions.HTTPError as error:
                   indigo.server.log("Healthy Home device creation error: "+self.getErrorText(error.response.text), isError=True)

            if self.weather:
                payload = {'access_token': self.accessToken}
                try:
                    response = requests.post("https://api.netatmo.com/api/getstationsdata", params=payload)
                    response.raise_for_status()
                    dataJ = response.json()["body"]
                    max_stations = len(dataJ["devices"])
                    for station in range(0, max_stations):
                        stnName = str(dataJ["devices"][station]["station_name"])
                        modName = str(dataJ["devices"][station]["module_name"])
                        stnAdd = "WS-"+str(station)
                        self.lastCO2[stnAdd] = 0
                        device = None
                        for dev in indigo.devices.iter("self"):
                            if dev.address == stnAdd:
                                device = indigo.devices[dev.name]
                        if device == None:
                            indigo.server.log("Creating Netatmo weather device for %s" % (modName))
                            nameM = "%s Air Monitor" % (modName)
                            device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=stnAdd, name=nameM, pluginId="com.racarter.indigoplugin.netatmo-welcome", deviceTypeId="netatmoBaseStation", props={})
                            device.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)
                        max_modules = len(dataJ["devices"][station]["modules"])
                        for module in range(0, max_modules):
                            modType = dataJ["devices"][station]["modules"][module]["type"]
                            stnName = str(dataJ["devices"][station]["station_name"])
                            modName = str(dataJ["devices"][station]["modules"][module]["module_name"])
                            modAdd = "WSM-"+str(module + 1)
                            if modType == "NAModule1":
                                #Outdoor Module
                                device = None
                                for dev in indigo.devices.iter("self"):
                                    if dev.address == modAdd:
                                        device = indigo.devices[dev.name]
                                if device == None:
                                    indigo.server.log("Creating Netatmo weather device for %s" % (modName))
                                    nameM = "%s Air Monitor" % (modName)
                                    device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=modAdd, name=nameM, pluginId="com.racarter.indigoplugin.netatmo-welcome", deviceTypeId="netatmoOutModule", props={})
                                    device.updateStateImageOnServer(indigo.kStateImageSel.TemperatureSensor)
                            elif modType == "NAModule2":
                                #Wind Module
                                device = None
                                for dev in indigo.devices.iter("self"):
                                    if dev.address == modAdd:
                                        device = indigo.devices[dev.name]
                                if device == None:
                                    indigo.server.log("Creating Netatmo weather device for %s" % (modName))
                                    nameM = "%s Wind" % (modName)
                                    device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=modAdd, name=nameM, pluginId="com.racarter.indigoplugin.netatmo-welcome", deviceTypeId="netatmoWindGauge", props={})
                                    device.updateStateImageOnServer(indigo.kStateImageSel.FanOff)
                            elif modType == "NAModule3":
                                #Rain Module
                                device = None
                                for dev in indigo.devices.iter("self"):
                                    if dev.address == modAdd:
                                        device = indigo.devices[dev.name]
                                if device == None:
                                    indigo.server.log("Creating Netatmo weather device for %s" % (modName))
                                    nameM = "%s Rainfall" % (modName)
                                    device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=modAdd, name=nameM, pluginId="com.racarter.indigoplugin.netatmo-welcome", deviceTypeId="netatmoRainGauge", props={})
                                    device.updateStateImageOnServer(indigo.kStateImageSel.HumiditySensor)
                            elif modType == "NAModule4":
                                #Additional Indoor Module
                                self.lastCO2[modAdd] = 0
                                device = None
                                for dev in indigo.devices.iter("self"):
                                    if dev.address == modAdd:
                                        device = indigo.devices[dev.name]
                                if device == None:
                                    indigo.server.log("Creating Netatmo weather device for %s" % (modName))
                                    nameM = "%s Air Monitor" % (modName)
                                    device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=modAdd, name=nameM, pluginId="com.racarter.indigoplugin.netatmo-welcome", deviceTypeId="netatmoAddIndoorModule", props={})
                                    device.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)
                            else:
                                indigo.server.log("Cannot create device: unknown module type %s" % modType, isError=True)
                                
                except requests.exceptions.HTTPError as error:
                    indigo.server.log("Weather device creation error: "+self.getErrorText(error.response.text), isError=True)


    def getUpdate(self):

        if self.accessToken == "" or self.refreshToken == "":
            status, detail = self.authenticate(self.userName, self.userPass, self.clientId, self.clientSecret)
            if status == 'error' and self.commsOK:
                indigo.server.log("Authentication failed in getUpdate():", isError=True)
                indigo.server.log(detail, isError=True)
                self.commsOK = False
                #return
            elif status == "OK" and self.commsOK == False:
                indigo.server.log("Authentication in getUpdate() now OK")
                self.commsOK = True

        if self.commsOK:
            if self.welcome:      
                payload = {'access_token': self.accessToken, 'size': 5,}
                try:
                    response = requests.post("https://api.netatmo.com/api/gethomedata", data=payload)
                    response.raise_for_status()
                    self.updateIndigoCam(response.json()["body"])
                    self.retriesC = 0        
                except:
                    self.retriesC += 1
                    if self.retriesC == 5:
                        indigo.server.log("Camera update error after five attempts: ", isError=True)
                        #self.commsOK = False
                    #self.sleep(7)
                
            if self.healthyHome and (time.time() - self.lastUpdateTimeHH > 300):
                payload = {'access_token': self.accessToken}
                try:
                    response = requests.post("https://api.netatmo.com/api/gethomecoachsdata", params=payload)
                    response.raise_for_status()
                    dataJ = response.json()["body"]
                    self.updateIndigoMon(response.json()["body"])
                    self.retriesH = 0
                except:
                    self.retriesH += 1
                    if self.retriesH == 5:
                        indigo.server.log("Healthy Home update error after five attempts", isError=True)
                        #self.commsOK = False
                    #self.sleep(7)
            
            if self.weather and (time.time() - self.lastUpdateTimeWS > 600):
                # Fresh readings should now be available.  (Weather station only uploads data every 10 minutes)
                payload = {'access_token': self.accessToken}
                try:
                    response = requests.post("https://api.netatmo.com/api/getstationsdata", params=payload)
                    response.raise_for_status()
                    dataJ = response.json()["body"]
                    self.updateIndigoWeather(response.json()["body"])
                    self.retriesW = 0
                except:
                    self.retriesW += 1
                    if self.retriesW == 5:
                        indigo.server.log("Weather update error after five attempts", isError=True)
                        #self.commsOK = False
                    #self.sleep(7)          
 
           
    def updateIndigoCam(self, dataJ):

        max_persons = len(dataJ["homes"][0]["persons"])
        home_count = 0
        self.currentPersonsList = []
        occ = "_Occupants"
        for pers in range(0, max_persons):
            try:
                try:
                    name = dataJ["homes"][0]["persons"][pers]["pseudo"]
                except:
                	# No value for 'pseudo' signifies person seen is unknown, so ignore when updating persons list.
                    continue
                name_id = dataJ["homes"][0]["persons"][pers]["id"]
                out = dataJ["homes"][0]["persons"][pers]["out_of_sight"]
                self.personId[name] = name_id
                if out:
                    status = "Away"
                else:
                    status = "Home"
                    home_count = home_count + 1
                    if self.loggingCam:
                        indigo.server.log(name+" is "+status)
                updateVar(name, value=status, folder=self.netatmoFolderId)
                self.currentPersonsList.append(name)
                if self.processLastSeen:
                    lastU = dataJ["homes"][0]["persons"][pers]["last_seen"]
                    if lastU > 0:
                        last = datetime.datetime.fromtimestamp(int(lastU)).strftime('%Y-%m-%d %H:%M:%S')
                    else:
                        last = "-"
                    nameLS = name+"_last_seen"
                    updateVar(nameLS, value=last, folder=self.netatmoFolderId)            
            except:
                e = sys.exc_info()[0]
                indigo.server.log("Cannot process %s, %s" % (name, e), isError=True)

        if self.delForgotten:
            self.delForgottenPersons()
            
        if occ not in indigo.variables:
            indigo.variable.create(occ, value=str(home_count), folder=self.netatmoFolderId)
        else:
            indigo.variable.updateValue(occ, value=str(home_count))

        max_cameras = len(dataJ["homes"][0]["cameras"])
        for cam in range(0, max_cameras):
            try:
                camState = dataJ["homes"][0]["cameras"][cam]["status"]
                if camState == "disconnected":
                    camState = "off"
                sdState = dataJ["homes"][0]["cameras"][cam]["sd_status"]
                psuState = dataJ["homes"][0]["cameras"][cam]["alim_status"]
                camName = str(dataJ["homes"][0]["cameras"][cam]["name"])
                if camState == "on":
                    try:
                        self.camUrl[cam] = str(dataJ["homes"][0]["cameras"][cam]["vpn_url"])
                    except:
                        # URL may not be sent for a short time after camera is switched on.
                        pass
                camAdd = "NW%s" % cam
                lastEventRaw = dataJ["homes"][0]["events"][0]["message"]
                lastEventTimeRaw = dataJ["homes"][0]["events"][0]["time"]
                lastEvent = re.sub('<[^<]+?>', '', lastEventRaw)
                lastEventTime = time.strftime('%d-%b-%y %H:%M', time.localtime(lastEventTimeRaw))
                keyValueList1 = [
                    {'key':'status', 'value':camState},
                    {'key':'sd_status', 'value':sdState},
                    {'key':'psu_status', 'value':psuState},
                ]
                keyValueList2 = [
                    {'key':'status', 'value':camState},
                    {'key':'sd_status', 'value':sdState},
                    {'key':'psu_status', 'value':psuState},
                    {'key':'last_event', 'value':lastEvent},
                    {'key':'last_event_time', 'value':lastEventTime},
                ]
                device = None
                for dev in indigo.devices.iter("self"):
                    if dev.address == camAdd:
                        device = indigo.devices[dev.name]
                if device != None:
                    device.updateStatesOnServer(keyValueList1)
                    lastEventCameraId = dataJ["homes"][0]["events"][0]["camera_id"]
                    if self.camId[cam] == lastEventCameraId:
                        device.updateStatesOnServer(keyValueList2)
                    else:
                        device.updateStatesOnServer(keyValueList1)                              
                    if camState == "off":
                        device.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)
                    else:
                        device.updateStateImageOnServer(indigo.kStateImageSel.SensorOn)
                    if self.syncCamNames:
                        camName = dataJ["homes"][0]["cameras"][cam]["name"] +" Camera"
                        if camName != device.name:
                            try:
                                device.name = camName
                                device.replaceOnServer()
                            except:
                                indigo.server.log("Cannot update %s name" % device.name, isError=True)
            except:
                e = sys.exc_info()[0]
                indigo.server.log("Camera %s update error: %s" % (camAdd, e), isError=True)  

        if self.loggingCam:
            indigo.server.log("Camera Data:\n"+str(dataJ))


    def updateIndigoMon(self, dataJ):
    
        max_monitors = len(dataJ["devices"])
        for mon in range(0, max_monitors):
            try:
                monAdd = "HH%s" % mon
                dataK = dataJ["devices"][mon]
                try:
                    health = dataK["dashboard_data"]["health_idx"]
                except:
                    health = 5
                if health == 0:
                    status = "Healthy"
                elif health == 1:
                    status = "Fine"
                elif health == 2:
                    status = "Fair"
                elif health == 3:
                    status = "Poor"
                elif health == 4:
                    status = "Unhealthy"
                else:
                    status = "Unknown"
                # Temperature and pressure trends are not immediately available when
                #  a monitor is installed, and occasionally Noise data is missing,
                #  hence the 'try' tests.
                try:
                    dbg = "Noise"
                    noise = dataK["dashboard_data"]["Noise"]
                except:
                    noise = 0
                dbg = "Temperature"
                temperature = dataK["dashboard_data"]["Temperature"]
                dbg = "Temp Trend"
                try:
                    tempTrend = dataK["dashboard_data"]["temp_trend"]
                except:
                    tempTrend = ""
                dbg = "Pressure"
                pressure = dataK["dashboard_data"]["Pressure"]
                dbg = "Press Trend"
                try:
                    pressureTrend = dataK["dashboard_data"]["pressure_trend"]
                except:
                    pressureTrend = ""
                dbg = "Humidity"
                humidity = dataK["dashboard_data"]["Humidity"]
                dbg = "CO2"
                cO2 = dataK["dashboard_data"]["CO2"]
                dbg = "Min Temp"
                minTemp = dataK["dashboard_data"]["min_temp"]
                dbg = "Min Temp Date"
                minTempDate = dataK["dashboard_data"]["date_min_temp"]
                minTempDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(minTempDate))
                dbg = "Max Temp"
                maxTemp = dataK["dashboard_data"]["max_temp"]
                dbg = "Max Temp Date"
                maxTempDate = dataK["dashboard_data"]["date_max_temp"]
                maxTempDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(maxTempDate))
                dbg = "co2Cal"
                co2Cal = dataK["co2_calibrating"]
                dbg = "Firmware"
                firmware = dataK["firmware"]
                dbg = "Last Upgrade"
                lastUpgrade = dataK["last_upgrade"]
                lastUpgradeF = time.strftime('%d-%b-%y %H:%M', time.localtime(lastUpgrade))
                dbg = "WiFi"
                wifi = dataK["wifi_status"]
                dbg = "UTC"
                timestamp = dataK["dashboard_data"]["time_utc"]
                timestampF = time.strftime('%d-%b-%y %H:%M', time.localtime(timestamp))
                self.lastUpdateTimeHH = timestamp
                device = None
                for dev in indigo.devices.iter("self"):
                    if dev.address == monAdd:
                        device = indigo.devices[dev.name]
                if device != None:
                    device.updateStateOnServer(key="Temperature", value=temperature, uiValue=str(temperature)+" °C")
                    keyValueList = [
                        {'key':'Status', 'value':status},
                        {'key':'Health', 'value':health},
                        {'key':'Noise', 'value':noise},
                        {'key':'TemperatureTrend', 'value':tempTrend},
                        {'key':'Pressure', 'value':pressure},
                        {'key':'PressureTrend', 'value':pressureTrend},
                        {'key':'Humidity', 'value':humidity},
                        {'key':'CO2', 'value':cO2},
                        {'key':'CO2Calibrating', 'value':co2Cal},
                        {'key':'MinTemp', 'value':minTemp},
                        {'key':'MinTempDate', 'value':minTempDateF},
                        {'key':'MaxTemp', 'value':maxTemp},
                        {'key':'MaxTempDate', 'value':maxTempDateF},
                        {'key':'Firmware', 'value':firmware},
                        {'key':'LastUpgrade', 'value':lastUpgradeF},
                        {'key':'WiFiStatus', 'value':wifi},
                        {'key':'LastReading', 'value':timestampF},
                    ]
                    device.updateStatesOnServer(keyValueList)
                    if health > 2:
                        device.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)
                    else:
                        device.updateStateImageOnServer(indigo.kStateImageSel.SensorOn)
                    if self.syncMonNames:
                        monName = dataJ["devices"][mon]["name"]+" Monitor"
                        if monName != device.name:
                            try:
                                device.name = monName
                                device.replaceOnServer()
                            except:
                                indigo.server.log("Cannot update %s name" % device.name, isError=True)                 

            except requests.exceptions.HTTPError as error:
                indigo.server.log("Healthy Home device update error: "+self.getErrorText(error.response.text), isError=True)
            except:
                e = sys.exc_info()[0]
                indigo.server.log("Healthy Home %s update error: %s, %s" % (monAdd, e, dbg), isError=True)

        if self.loggingMon:
            indigo.server.log("Healthy Home Data:\n"+str(dataJ))
            

    def updateIndigoWeather(self, dataJ):
    
        max_stations = len(dataJ["devices"])
        for station in range(0, max_stations):
            try:
                dataK = dataJ["devices"][station]
                dataL = dataJ["user"]["administrative"]
                modType = dataK["type"]
                stnName = str(dataK["station_name"])
                modName = str(dataK["module_name"])
                stnAdd = "WS-"+str(station)
                # Base Station (noise, temp, mintemp+date, maxtemp+date, humidity, pressure, co2, wifi, timestamp)
                # Temperature and pressure trends are not immediately available when
                #  a monitor is installed, and occasionally Noise data is missing,
                #  hence the 'try' tests.
                try:
                    noise = dataK["dashboard_data"]["Noise"]
                except:
                    noise = 0
                temperature = dataK["dashboard_data"]["Temperature"]
                try:
                    tempTrend = dataK["dashboard_data"]["temp_trend"]
                except:
                    tempTrend = ""
                pressure = dataK["dashboard_data"]["Pressure"]
                absPressure = dataK["dashboard_data"]["AbsolutePressure"]
                humidity = dataK["dashboard_data"]["Humidity"]
                try:
                    cO2 = dataK["dashboard_data"]["CO2"]
                    self.lastCO2[stnAdd] = cO2
                except:
                    cO2 = self.lastCO2[stnAdd]
                co2Cal = dataK["co2_calibrating"]
                minTemp = dataK["dashboard_data"]["min_temp"]
                minTempDate = dataK["dashboard_data"]["date_min_temp"]
                minTempDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(minTempDate))
                maxTemp = dataK["dashboard_data"]["max_temp"]
                maxTempDate = dataK["dashboard_data"]["date_max_temp"]
                maxTempDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(maxTempDate))
                firmware = dataK["firmware"]
                lastUpgrade = dataK["last_upgrade"]
                lastUpgradeF = time.strftime('%d-%b-%y %H:%M', time.localtime(lastUpgrade))
                wifi = dataK["wifi_status"]
                tUnit = self.tUnits[dataL["unit"]]
                rUnit = self.rUnits[dataL["unit"]]
                wUnit = self.wUnits[dataL["windunit"]]
                pUnit = self.pUnits[dataL["pressureunit"]]
                wCF = self.wConversion[dataL["windunit"]]
                timestamp = dataK["dashboard_data"]["time_utc"]
                timestampF = time.strftime('%d-%b-%y %H:%M', time.localtime(timestamp))
                self.lastUpdateTimeWS = timestamp
                device = None
                for dev in indigo.devices.iter("self"):
                    if dev.address == stnAdd:
                        device = indigo.devices[dev.name]
                if device != None:
                    device.updateStateOnServer(key="Temperature", value=temperature, uiValue=str(temperature)+tUnit)
                    keyValueList = [
                        {'key':'Noise', 'value':noise},
                        {'key':'TemperatureTrend', 'value':tempTrend},
                        {'key':'Pressure', 'value':pressure},
                        {'key':'AbsolutePressure', 'value':absPressure},
                        {'key':'Humidity', 'value':humidity},
                        {'key':'CO2', 'value':cO2},
                        {'key':'CO2Calibrating', 'value':co2Cal},
                        {'key':'MinTemp', 'value':minTemp},
                        {'key':'MinTempDate', 'value':minTempDateF},
                        {'key':'MaxTemp', 'value':maxTemp},
                        {'key':'MaxTempDate', 'value':maxTempDateF},
                        {'key':'Firmware', 'value':firmware},
                        {'key':'LastUpgrade', 'value':lastUpgradeF},
                        {'key':'WiFiStatus', 'value':wifi},
                        {'key':'LastReading', 'value':timestampF},
                    ]
                    device.updateStatesOnServer(keyValueList)
                    if cO2 > 1200:
                        device.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)
                    else:
                        device.updateStateImageOnServer(indigo.kStateImageSel.SensorOn)

                # Process Modules:
                # NAModule1 = Outdoor (temp, mintemp+date, maxtemp+date, humidity, pressure, ptrend, radio, battery, timestamp)
                # NAModule2 = Wind (strength, angle, gust strength, gust angle, last hour, radio, battery, timestamp)
                # NAModule3 = Rain (past hour, past 24h, radio, battery, timestamp)
                # NAModule4 = Additional Indoor (as per base station - but no noise or pressure, and radio instead of wifi)
                max_modules = len(dataJ["devices"][station]["modules"])
                for module in range(0, max_modules):
                    stnName = str(dataJ["devices"][station]["station_name"])
                    dataK = dataJ["devices"][station]["modules"][module]
                    modType = dataK["type"]
                    modName = str(dataK["module_name"])
                    #Debug code to list dictionary items/values to Event Log:
                    #for k, v in dataK.iteritems():
                    #   indigo.server.log(str(v), str(k))
                    modAdd = "WSM-"+str(module + 1)
                    if modType == "NAModule1":
                        # Outdoor Module
                        try:
                            temperature = dataK["dashboard_data"]["Temperature"]
                            try:
                                tempTrend = dataK["dashboard_data"]["temp_trend"]
                            except:
                                tempTrend = ""
                            humidity = dataK["dashboard_data"]["Humidity"]
                            minTemp = dataK["dashboard_data"]["min_temp"]
                            minTempDate = dataK["dashboard_data"]["date_min_temp"]
                            minTempDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(minTempDate))
                            maxTemp = dataK["dashboard_data"]["max_temp"]
                            maxTempDate = dataK["dashboard_data"]["date_max_temp"]
                            maxTempDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(maxTempDate))
                            firmware = dataK["firmware"]
                            radio = dataK["rf_status"]
                            battery = dataK["battery_percent"]
                            timestamp = dataK["dashboard_data"]["time_utc"]
                            timestampF = time.strftime('%d-%b-%y %H:%M', time.localtime(timestamp))
                            device = None
                            for dev in indigo.devices.iter("self"):
                                if dev.address == modAdd:
                                    device = indigo.devices[dev.name]
                            if device != None:
                                device.updateStateOnServer(key="Temperature", value=temperature, uiValue=str(temperature)+tUnit)
                                keyValueList = [
                                    {'key':'TemperatureTrend', 'value':tempTrend},
                                    {'key':'Humidity', 'value':humidity},
                                    {'key':'MinTemp', 'value':minTemp},
                                    {'key':'MinTempDate', 'value':minTempDateF},
                                    {'key':'MaxTemp', 'value':maxTemp},
                                    {'key':'MaxTempDate', 'value':maxTempDateF},
                                    {'key':'batteryLevel', 'value':battery},
                                    {'key':'Firmware', 'value':firmware},
                                    {'key':'RadioStatus', 'value':radio},
                                    {'key':'LastReading', 'value':timestampF},
                                ]
                                device.updateStatesOnServer(keyValueList)
                        except:
                            e = sys.exc_info()[0]
                            indigo.server.log("Outdoor module %s update error: %s" % (modAdd, e), isError=True)

                    elif modType == "NAModule2":
                        # Wind Gauge
                        try:
                            dbg = "Strength"
                            if wCF > 0:
                                wStrength = int(round(dataK["dashboard_data"]["WindStrength"] * wCF))
                            else:
                                # Beaufort Scale selected
                                wStrength = self.kphToBft(dataK["dashboard_data"]["WindStrength"])
                            dbg = "Wangle"
                            wAngle = dataK["dashboard_data"]["WindAngle"]
                            wDirection = self.degToCompass(wAngle)
                            dbg = "Gstrength"
                            gStrength = int(round(dataK["dashboard_data"]["GustStrength"] * wCF))
                            dbg = "Gangle"
                            gAngle = dataK["dashboard_data"]["GustAngle"]
                            gDirection = self.degToCompass(gAngle)
                            dbg = "Mstrength"
                            wMax = int(round(dataK["dashboard_data"]["max_wind_str"] * wCF))
                            dbg = "Mdate"
                            wMaxDate = dataK["dashboard_data"]["date_max_wind_str"]
                            wMaxDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(wMaxDate))
                            dbg = "Firmware"
                            firmware = dataK["firmware"]
                            dbg = "Radio"
                            radio = dataK["rf_status"]
                            dbg = "Battery"
                            battery = dataK["battery_percent"]
                            dbg = "UTC"
                            timestamp = dataK["dashboard_data"]["time_utc"]
                            timestampF = time.strftime('%d-%b-%y %H:%M', time.localtime(timestamp))
                            device = None
                            for dev in indigo.devices.iter("self"):
                                if dev.address == modAdd:
                                    device = indigo.devices[dev.name]
                            if device != None:
                                device.updateStateOnServer(key="WindStrength", value=wStrength, uiValue=str(wStrength)+wUnit)
                                keyValueList = [
                                    {'key':'WindAngle', 'value':wAngle},
                                    {'key':'WindDirection', 'value':wDirection},
                                    {'key':'MaxWindStrength', 'value':wMax},
                                    {'key':'MaxWindDate', 'value':wMaxDateF},
                                    {'key':'GustStrength', 'value':gStrength},
                                    {'key':'GustAngle', 'value':gAngle},
                                    {'key':'GustDirection', 'value':gDirection},
                                    {'key':'batteryLevel', 'value':battery},
                                    {'key':'Firmware', 'value':firmware},
                                    {'key':'RadioStatus', 'value':radio},
                                    {'key':'LastReading', 'value':timestampF},
                                ]
                                device.updateStatesOnServer(keyValueList)
                                if self.showDirectionInNotes:
                                    device.description = wDirection
                                    device.replaceOnServer()
                        except:
                            e = sys.exc_info()[0]
                            indigo.server.log("Wind gauge %s update error: %s, %s" % (modAdd, e, dbg), isError=True)
    
                    elif modType == "NAModule3":
                        # Rain Gauge
                        try:
                            try:
                                rain = dataK["dashboard_data"]["Rain"]
                                self.lastRain = rain
                            except:
                                rain = self.lastRain
                            try:
                                rainH = round(dataK["dashboard_data"]["sum_rain_1"], 1)
                                self.lastRainH = rainH
                            except:
                                rainH = self.lastRainH
                            try:
                                rainD = round(dataK["dashboard_data"]["sum_rain_24"], 1)
                                self.lastRainD = rainD
                            except:
                                rainD = self.lastRainD
                            if dataL["unit"] == 1:
                                #Convert mm to inches
                                rain = round((rain * 0.0393701), 2)
                                rainH = round((rainH * 0.0393701), 2)
                                rainD = round((rainD * 0.0393701), 2)
                            firmware = dataK["firmware"]
                            radio = dataK["rf_status"]
                            battery = dataK["battery_percent"]
                            timestamp = dataK["dashboard_data"]["time_utc"]
                            timestampF = time.strftime('%d-%b-%y %H:%M', time.localtime(timestamp))
                            device = None
                            for dev in indigo.devices.iter("self"):
                                if dev.address == modAdd:
                                    device = indigo.devices[dev.name]
                            if device != None:
                                device.updateStateOnServer(key="Rain", value=rain, uiValue=str(rain)+rUnit)
                                keyValueList = [
                                    {'key':'RainLastHour', 'value':rainH},
                                    {'key':'RainLastDay', 'value':rainD},
                                    {'key':'batteryLevel', 'value':battery},
                                    {'key':'Firmware', 'value':firmware},
                                    {'key':'RadioStatus', 'value':radio},
                                    {'key':'LastReading', 'value':timestampF},
                                ]
                                device.updateStatesOnServer(keyValueList)
                        except:
                            e = sys.exc_info()[0]
                            indigo.server.log("Rain gauge %s update error: %s" % (modAdd, e), isError=True)                       
                        
                    elif modType == "NAModule4":
                        # Additional Indoor Module
                        try:
                            temperature = dataK["dashboard_data"]["Temperature"]
                            try:
                                tempTrend = dataK["dashboard_data"]["temp_trend"]
                            except:
                                tempTrend = ""
                            humidity = dataK["dashboard_data"]["Humidity"]
                            try:
                                cO2 = dataK["dashboard_data"]["CO2"]
                                self.lastCO2[modAdd] = cO2
                            except:
                                cO2 = self.lastCO2[modAdd]
                            minTemp = dataK["dashboard_data"]["min_temp"]
                            minTempDate = dataK["dashboard_data"]["date_min_temp"]
                            minTempDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(minTempDate))
                            maxTemp = dataK["dashboard_data"]["max_temp"]
                            maxTempDate = dataK["dashboard_data"]["date_max_temp"]
                            maxTempDateF = time.strftime('%d-%b-%y %H:%M', time.localtime(maxTempDate))
                            firmware = dataK["firmware"]
                            radio = dataK["rf_status"]
                            battery = dataK["battery_percent"]
                            timestamp = dataK["dashboard_data"]["time_utc"]
                            timestampF = time.strftime('%d-%b-%y %H:%M', time.localtime(timestamp))
                            device = None
                            for dev in indigo.devices.iter("self"):
                                if dev.address == modAdd:
                                    device = indigo.devices[dev.name]
                            if device != None:
                                # ***This code was required to fix some devices not showing battery level bar
                                # localPropsCopy = device.pluginProps
								# localPropsCopy.update({"SupportsBatteryLevel":True})
                        		# device.replacePluginPropsOnServer(localPropsCopy)
                                # ***
                                device.updateStateOnServer(key="Temperature", value=temperature, uiValue=str(temperature)+tUnit)
                                keyValueList = [
                                    {'key':'TemperatureTrend', 'value':tempTrend},
                                    {'key':'Humidity', 'value':humidity},
                                    {'key':'CO2', 'value':cO2},
                                    {'key':'MinTemp', 'value':minTemp},
                                    {'key':'MinTempDate', 'value':minTempDateF},
                                    {'key':'MaxTemp', 'value':maxTemp},
                                    {'key':'MaxTempDate', 'value':maxTempDateF},
                                    {'key':'batteryLevel', 'value':battery},
                                    {'key':'Firmware', 'value':firmware},
                                    {'key':'RadioStatus', 'value':radio},
                                    {'key':'LastReading', 'value':timestampF},
                                ]
                                device.updateStatesOnServer(keyValueList)
                                if cO2 > 1200:
                                    device.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)
                                else:
                                    device.updateStateImageOnServer(indigo.kStateImageSel.SensorOn)
                        except:
                            e = sys.exc_info()[0]
                            indigo.server.log("Indoor module %s update error: %s" % (modAdd, e), isError=True)
                      
                    else:
                        indigo.server.log("Weather update: unknown module type %s" % modType, isError=True)
                        
            except requests.exceptions.HTTPError as error:
                indigo.server.log("Weather Station update error: "+self.getErrorText(error.response.text), isError=True)
            except:
                e = sys.exc_info()[0]
                indigo.server.log("Weather station %s update error: %s" % (stnAdd, e), isError=True)

        if self.loggingWs:
            indigo.server.log("Weather Station Data:\n"+str(dataJ)) 
            
            
    def degToCompass(self, num):
        #Convert degrees to compass directions
        val=int((num/22.5) + 0.5)
        arr=["N","NNE","NE","ENE","E","ESE", "SE", "SSE","S","SSW","SW","WSW","W","WNW","NW","NNW"]
        return arr[(val % 16)]


    def kphToBft(self, kph):
        #Convert wind from kph to Beaufort scale
        if kph is None:
            return None
        for bft in range(len(self.bftThreshold)):
            if kph < self.bftThreshold[bft]:
                return bft
        return len(self.bftThreshold)


    def getErrorText(self, html):
        #Extract error text from html error message
        i = html.find("<title>")
        if i != -1:
            j = html.find("</title>")
            return html[i+7:j]
        else:
            return "Unknown error"

    
    def authenticate(self, userName, userPass, clientId, clientSecret):
        #Obtain correct authentication for selected device types
        if self.loggingCam or self.loggingMon or self.loggingWs:
            indigo.server.log("Authenticating...")
            
        if self.welcome and self.healthyHome and not self.weather:
            if self.loggingCam or self.loggingMon:
                indigo.server.log("Authenticating for Welcome Camera and Healthy Home")
            payload = {'grant_type': 'password',
            	'client_id': clientId,
            	'client_secret': clientSecret,
                'username': userName,
                'password': userPass,
                'scope': 'read_camera write_camera access_camera read_homecoach'}
                
        elif self.welcome and self.weather and not self.healthyHome:
            if self.loggingCam or self.loggingWs:
                indigo.server.log("Authenticating for Welcome Camera and Weather Station")
            payload = {'grant_type': 'password',
            	'client_id': clientId,
            	'client_secret': clientSecret,
                'username': userName,
                'password': userPass,
                'scope': 'read_camera write_camera access_camera read_station'}
                
        elif self.welcome and self.healthyHome and self.weather:
            if self.loggingCam or self.loggingMon or self.loggingWs:
                indigo.server.log("Authenticating for Welcome Camera, Healthy Home and Weather Station")
            payload = {'grant_type': 'password',
            	'client_id': clientId,
            	'client_secret': clientSecret,
                'username': userName,
                'password': userPass,
                'scope': 'read_camera write_camera access_camera read_homecoach read_station'}
                
        elif not self.welcome and self.healthyHome and self.weather:
            if self.loggingMon or self.loggingWs:
                indigo.server.log("Authenticating for Healthy Home and Weather Station")
            payload = {'grant_type': 'password',
            	'client_id': clientId,
            	'client_secret': clientSecret,
                'username': userName,
                'password': userPass,
                'scope': 'read_homecoach read_station'}

        elif not self.welcome and not self.healthyHome and self.weather:
            if self.loggingWs:
                indigo.server.log("Authenticating for Weather Station only")
            payload = {'grant_type': 'password',
            	'client_id': clientId,
            	'client_secret': clientSecret,
                'username': userName,
                'password': userPass,
                'scope': 'read_station'}
                
        elif not self.welcome and self.healthyHome and not self.weather:
            if self.loggingMon:
                indigo.server.log("Authenticating for Healthy Home only")
            payload = {'grant_type': 'password',
            	'client_id': clientId,
            	'client_secret': clientSecret,
                'username': userName,
                'password': userPass,
                'scope': 'read_homecoach'}

        elif self.welcome and not self.healthyHome and not self.weather:
            if self.loggingCam:
                indigo.server.log("Authenticating for Welcome Camera only")
            payload = {'grant_type': 'password',
            	'client_id': clientId,
            	'client_secret': clientSecret,
                'username': userName,
                'password': userPass,
                'scope': 'read_camera write_camera access_camera'}
                
        try:
            response = requests.post("https://api.netatmo.com/oauth2/token", data=payload)
            response.raise_for_status()
            self.accessToken=response.json()["access_token"]
            self.refreshToken=response.json()["refresh_token"]
            self.expiration = int(response.json()['expire_in'] + time.time())
            self.pluginPrefs["accessToken"] = self.accessToken
            self.pluginPrefs["refreshToken"] = self.refreshToken
            self.pluginPrefs["expiration"] = self.expiration
            if self.loggingCam or self.loggingMon or self.loggingWs:
                indigo.server.log("Your access token is: "+str(self.accessToken))
                indigo.server.log("Your refresh token is: "+str(self.refreshToken))
            return "OK", response.status_code 
        #except requests.exceptions.ConnectionError:
            #indigo.server.log("Authenticate error - cannot connect to Netatmo servers", isError=True)
        #   self.commsOK = False
        #   return "error"
        except requests.exceptions.RequestException as e:
            return "error", str(e)

         
    def refresh(self, clientId, clientSecret, refreshToken):
        #Refresh Netatmo token on expiry
        if self.loggingCam or self.loggingMon or self.loggingWs:
            indigo.server.log("Refreshing Netatmo token...")
            
        payload = {'grant_type': 'refresh_token',
           'refresh_token': refreshToken,
           'client_id': clientId,
           'client_secret': clientSecret,}
        try:
            response = requests.post("https://api.netatmo.com/oauth2/token", data=payload)
            response.raise_for_status()
            self.accessToken=response.json()["access_token"]
            self.refreshToken=response.json()["refresh_token"]
            self.expiration = int(response.json()['expire_in'] + time.time())
            self.pluginPrefs["accessToken"] = self.accessToken
            self.pluginPrefs["refreshToken"] = self.refreshToken
            self.pluginPrefs["expiration"] = self.expiration
            if self.loggingCam or self.loggingMon or self.loggingWs:
                indigo.server.log("Your access token is: "+str(self.accessToken))
                indigo.server.log("Your refresh token is: "+str(self.refreshToken))
            return "OK", response.status_code
        #except requests.exceptions.ConnectionError:
            #indigo.server.log("Refresh error - cannot connect to Netatmo servers", isError=True)
        #   return "error"
        except requests.exceptions.RequestException as e:
            self.authenticate(self.userName, self.userPass, self.clientId, self.clientSecret)
            return "error", str(e)
            
            
    def delForgottenPersons(self):
    
        #Checks for person and Last Seen variables in Netatmo folder which are no longer
        # being reported by Netatmo, and deletes them if they have no dependencies.  If
        # dependencies exist an error message and details are published in the Event Log.
        personsDeleted = 0
        for var in indigo.variables.iter("self"):
            if (var.folderId == self.netatmoFolderId) and ("_Occupants" not in var.name) and ("_last_seen" not in var.name):
                if var.name not in self.currentPersonsList:
                    indigo.server.log(var.name+" not in current Netatmo persons list")
                    depsList = indigo.variable.getDependencies(var)
                    #indigo.server.log(str(depsList))
                    if any (depsList.itervalues()):
                        lenAct = len(depsList["actionGroups"])
                        lenCon = len(depsList["controlPages"])
                        lenDev = len(depsList["devices"])
                        lenSch = len(depsList["schedules"])
                        lenTri = len(depsList["triggers"])
                        lenVar = len(depsList["variables"])
                        indigo.server.log("Variable "+var.name+" has dependencies - cannot delete", isError=True)
                        indigo.server.log("%d action group dependencies" % lenAct)
                        indigo.server.log("%d control page dependencies" % lenCon)
                        indigo.server.log("%d device dependencies" % lenDev)
                        indigo.server.log("%d schedule dependencies" % lenSch)
                        indigo.server.log("%d trigger dependencies" % lenTri)
                        indigo.server.log("%d variable dependencies" % lenVar)
                    else:
                        indigo.variable.delete(var)
                        indigo.server.log("Variable "+var.name+" deleted")
                        personsDeleted = personsDeleted + 1    
                        vLastSeen = var.name+"_last_seen"
                        if  (vLastSeen in indigo.variables):
                            indigo.variable.delete(vLastSeen)
                            indigo.server.log("Variable "+vLastSeen+" deleted")
        return personsDeleted
        

        
    ########################################
    # Menu Item functions
    ######################
    
    def updateReadingsNow(self):
    
        #Ad hoc poll for latest data
        self.lastUpdateTimeHH = 0
        self.lastUpdateTimeWS = 0
        self.getUpdate() 
        indigo.server.log("Manual update of readings is complete")
    

    def checkVersionNow(self):
    
        #Calls the Indigo Plugin Update Checker in response to a user request
        try:
            self.updater.checkVersionNow()
        except Exception as error:
            self.errorLog(u"Error checking plugin update status. Error: (Line {0}  {1})".format(sys.exc_traceback.tb_lineno, error))
            return False
    
    
    def removeDeletedPersons(self):
    
        #Deletes person and last seen variables for people not in current Netatmo reply
        if self.welcome:   
            indigo.server.log("Checking for deleted persons...")
            numberDeleted = self.delForgottenPersons()
            if numberDeleted == 0:
                indigo.server.log("All persons are in the current list; no variables deleted")
        else:
            indigo.server.log("Invalid request - no Welcome camera present")
    
    
    #######################################
    # Action functions
    ######################
    
    def setPersonAway(self, action):
    
        if self.welcome:
            name = action.props.get("name", "")
            if name in self.personId:
                payload = {'access_token': self.accessToken, 'home_id': self.homeId, 'person_id': self.personId[name], 'size': 2,}
                try:
                    response = requests.post("https://api.netatmo.com/api/setpersonsaway", data=payload)
                    response.raise_for_status()
                    indigo.server.log(name+" set away")
                except:
                    indigo.server.log("Set Person Away - server access error", isError=True)
            else:
                indigo.server.log("Set Person Away - "+name+" not found", isError=True)
        else:
            indigo.server.log("Invalid request - no Welcome camera present")

            
    def setAllAway(self, action):
    
        if self.welcome:
            payload = {'access_token': self.accessToken, 'home_id': self.homeId, 'size': 2,}
            try:
                response = requests.post("https://api.netatmo.com/api/setpersonsaway", data=payload)
                response.raise_for_status()
                indigo.server.log("All persons set away")
            except:
                indigo.server.log("Set All Away - server access error", isError=True)
        else:
            indigo.server.log("Invalid request - no Welcome camera present")


    def setPersonHome(self, action):
    
        if self.welcome:
            name = action.props.get("name", "")
            if name in self.personId:
#                 pArray = "{"+self.personId[name]+"}"
                payload = {'access_token': self.accessToken, 'home_id': self.homeId, 'person_ids[]': self.personId[name],}
                try:
                	response = requests.post("https://api.netatmo.com/api/setpersonshome", data=payload)
                	response.raise_for_status()
                	indigo.server.log(name+" set home")
                except:
                    indigo.server.log("Set Person Home - server access error", isError=True)
            else:
                indigo.server.log("Set Person Home - "+name+" not found", isError=True)
        else:
            indigo.server.log("Invalid request - no Welcome camera present")


    def takeSnapshot(self, action):

        errorDict = indigo.Dict()
        try:
            devId = action.props.get("devId", "")
            if devId == "":
                indigo.server.log("Please specify a camera device for snapshot", isError=True)
                return False
            if self.snapshotPath == "":
                indigo.server.log("Please specify a folder path for snapshot", isError=True)
                return False
            device = indigo.devices[int(devId)]
            cam = device.address[2:]
            if self.camUrl[int(cam)] != "":
                imgLoc = self.snapshotPath+"/welcome%s %s.jpg" % (cam, datetime.datetime.now().strftime('%Y-%m-%d %H-%M-%S'))
                request = urllib2.Request(self.camUrl[int(cam)]+"/live/snapshot_720.jpg") 
                buffer = urllib2.urlopen(request).read()
                f = open(imgLoc, 'wb')
                f.write(buffer)
                f.close()
                return True
            else:
                indigo.server.log("Missing URL for camera snapshot", isError=True)
        except:
            indigo.server.log("Snapshot error", isError=True)
            return False
